import io
import sys
import unittest

from argparse import Namespace
from contextlib import contextmanager

from ticket_universe import cli


@contextmanager
def capture(command, *args, **kwargs):
    out, sys.stdout = sys.stdout, io.StringIO()
    try:
        command(*args, **kwargs)
        sys.stdout.seek(0)
        yield sys.stdout.read()
    finally:
        sys.stdout = out


class CliTest(unittest.TestCase):
    def test_limiting_output(self):
        args = Namespace(positions=["binary", "binary", "binary"], limit=4, offset=0)
        with capture(cli.main, args) as output:
            self.assertEqual(4, output.count('\n'))

    def test_offsetting_output(self):
        args = Namespace(positions=["binary", "binary", "binary"], limit=None, offset=4)
        with capture(cli.main, args) as output:
            self.assertEqual(4, output.count('\n'))

    def test_calling_main_without_args_prints_help(self):
        args = cli.arguments([])
        with capture(cli.main, args) as o, capture(cli.print_help) as h:
            self.assertEqual(h, o)

    def test_calling_main_with_args_prints_universe(self):
        args = Namespace(positions=["binary"], limit=None, offset=0)
        with capture(cli.main, args) as output:
            self.assertEqual("0\n1\n", output)



from typing import Union

from ticket_universe.errors import MissingCharset


class Position:
    def __init__(self, characters: [str]):
        self.characters = characters

    def __len__(self):
        return len(self.characters)

    def __iter__(self):
        return iter(self.characters)

    @classmethod
    def from_string(cls, s) -> "Position":
        chunks = s.split(":")
        _type = chunks[0]
        _args = chunks[1:]
        return build_type(_type, _args)


class FixedPosition(Position):
    def __init__(self, character: str):
        Position.__init__(self, [character])


class RangedPosition(Position):
    def __init__(self, _min: Union[str, int], _max: Union[str, int]):
        _range = (
            [str(num) for num in range(int(_min), int(_max) + 1)]
            if _min != _max
            else []
        )
        Position.__init__(self, _range)


class BinaryPosition(Position):
    def __init__(self):
        Position.__init__(self, ["0", "1"])


class NumericPosition(Position):
    def __init__(self):
        Position.__init__(self, list("0123456789"))


class AlphaPosition(Position):
    def __init__(self, charset: str = "latin"):
        Position.__init__(self, self._import_charset(charset))

    @staticmethod
    def _import_charset(charset) -> [str]:
        import ticket_universe.charsets as charset_factories

        charset_factory = getattr(charset_factories, charset, None)
        if charset_factory is None:
            raise MissingCharset(charset)

        return charset_factory()


def build_type(_type: str, args):
    import sys

    module = sys.modules[__name__]
    position_type = getattr(module, _type.title() + "Position")
    return position_type(*args)



import argparse

from ticket_universe import universe
from ticket_universe.position import Position


def arguments(*args, **kwargs) -> argparse.Namespace:
    parser = argument_parser()
    args = parser.parse_args(*args, **kwargs)
    return args


def argument_parser():
    command_desc = "example: ticket-universe fixed:LTFY- alpha numeric numeric numeric"
    positions_help = (
        "alpha | alpha:safe_latin | numeric | range:min:max | fixed:{} | binary"
    )
    parser = argparse.ArgumentParser("ticket-universe", description=command_desc)
    parser.add_argument(
        "positions", metavar="POSITION", type=str, nargs="*", help=positions_help
    )
    parser.add_argument("--offset", default=0, type=int)
    parser.add_argument("--limit", default=None, type=int)
    return parser


def universe_from_args(args):
    return universe.Universe([Position.from_string(s) for s in args.positions])


def main(args: argparse.Namespace = None):
    _arguments = args or arguments()
    _universe = universe_from_args(_arguments)

    if len(_universe) == 0:
        print_help()

    for t in _universe.generate(_arguments.limit, _arguments.offset):
        print(t)


def print_help():
    parser = argument_parser()
    parser.print_help()



import unittest

from ticket_universe import charsets, errors
from ticket_universe.position import (
    Position,
    AlphaPosition,
    FixedPosition,
    BinaryPosition,
    NumericPosition,
    RangedPosition,
)


class PositionTest(unittest.TestCase):
    def test_position_length_equals_character_length(self):
        chars = list("123")
        self.assertEqual(len(chars), len(Position(chars)))

    def test_position_iteration_equals_character_list(self):
        chars = list("123")
        self.assertEqual(chars, [char for char in Position(chars)])

    def test_building_position_from_string(self):
        ranged_position = Position.from_string("ranged:0:255")
        self.assertIsInstance(ranged_position, RangedPosition)
        self.assertEqual(256, len(ranged_position))

        alpha_position = Position.from_string("alpha")
        self.assertIsInstance(alpha_position, AlphaPosition)
        self.assertEqual(charsets.latin(), alpha_position.characters)

        alpha_latin_safe_position = Position.from_string("alpha:safe_latin")
        self.assertIsInstance(alpha_latin_safe_position, AlphaPosition)
        self.assertEqual(charsets.safe_latin(), alpha_latin_safe_position.characters)

        expected_fixed_value = "fixed-element-"
        fixed_position = Position.from_string("fixed:" + expected_fixed_value)
        self.assertIsInstance(fixed_position, FixedPosition)
        self.assertEqual(1, len(fixed_position))
        self.assertEqual([expected_fixed_value], fixed_position.characters)


class FixedPositionTest(unittest.TestCase):
    def test_position_length_equals_one(self):
        self.assertEqual(1, len(FixedPosition("a")))
        self.assertEqual(1, len(FixedPosition("ab")))


class RangedPositionTest(unittest.TestCase):
    def test_range_expectations_match_length(self):
        self.assertEqual(256, len(RangedPosition(0, 255)))
        self.assertEqual(0, len(RangedPosition(0, 0)))


class BinaryPositionTest(unittest.TestCase):
    def test_position_length_equals_two(self):
        self.assertEqual(2, len(BinaryPosition()))


class NumericPositionTest(unittest.TestCase):
    def test_position_length_equals_ten(self):
        self.assertEqual(10, len(NumericPosition()))


class AlphaPositionTest(unittest.TestCase):
    def test_sane_defaults_are_used(self):
        self.assertTrue(len(AlphaPosition()) > 0)

    def test_charset_can_be_specified(self):
        self.assertEqual(
            charsets.safe_latin(), [char for char in AlphaPosition("safe_latin")]
        )

    def test_invalid_charset_raises_exception(self):
        with self.assertRaises(errors.MissingCharset):
            AlphaPosition("invalid_charset")



import unittest

from ticket_universe.position import BinaryPosition
from ticket_universe.universe import Universe


class UniverseTest(unittest.TestCase):
    def test_universe_size_equals_size_cartesian_product_size(self):
        universe = Universe([BinaryPosition(), BinaryPosition(), BinaryPosition()])
        tickets = [ticket for ticket in universe]
        self.assertEqual(
            8,
            len(tickets),
            "A binary universe with three positions should be length 2 ** 3 == 8",
        )
        self.assertEqual(
            len(universe),
            len(tickets),
            "Universe length should equal resulting ticket list length",
        )

    def test_universe_is_unique(self):
        universe = Universe([BinaryPosition(), BinaryPosition(), BinaryPosition()])
        tickets = [ticket for ticket in universe.generate()]
        self.assertEqual(8, len(set(tickets)), "The binary universe should be unique")

    def test_limited_universe(self):
        ticket_limit = 3
        universe = Universe([BinaryPosition(), BinaryPosition(), BinaryPosition()])
        tickets = [ticket for ticket in universe.generate(limit=ticket_limit)]
        self.assertEqual(
            ticket_limit, len(tickets), "The universe should obey the ticket limit"
        )

    def test_offset_universe(self):
        offset = 2
        universe = Universe([BinaryPosition(), BinaryPosition(), BinaryPosition()])
        tickets = [ticket for ticket in universe.generate(offset=offset)]
        self.assertEqual(6, len(tickets))

    def test_offset_limit_universe(self):
        universe = Universe([BinaryPosition(), BinaryPosition(), BinaryPosition()])
        tickets = [ticket for ticket in universe.generate(limit=4, offset=2)]
        self.assertEqual(2, len(tickets))



import functools
import itertools
import operator

from typing import Generator

from ticket_universe.position import Position


class Universe:
    def __init__(self, positions: [Position]):
        self.positions = positions

    def __iter__(self) -> Generator[str, None, None]:
        return self.generate()

    def __len__(self) -> int:
        """ Calculates total size of the universe as a product of the size of each position """
        if len(self.positions) == 0:
            return 0
        return functools.reduce(operator.mul, map(len, self.positions))

    def generate(self, limit=None, offset=0) -> Generator[str, None, None]:
        """ Generates unique ticket codes within the universe """
        characters_product = itertools.product(
            *self.positions
        )  # cartesian product of character groups

        _limit = limit or len(self)

        for i, characters in enumerate(characters_product):
            if i < offset:
                continue
            if i >= _limit:
                return
            yield "".join(characters)



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name='ticket_universe',
    version='0.4',
    description='Toolset to generate ticket universes',
    long_description=long_description,
    long_description_content_type="text/markdown",
    url='http://github.com/lotify/ticket_universe',
    author='Eelke van den Bos',
    author_email='eelke@moddix.com',
    license='MIT',
    packages=setuptools.find_packages(),
    entry_points={
        'console_scripts': ['ticket-universe=ticket_universe.cli:main']
    },
    zip_safe=False
)
